# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods that manage the AWS ApiGateway at high level."""

from typing import Dict
from scar.providers.aws import GenericClient
import scar.logger as logger

# 'HTTP'|'AWS'|'MOCK'|'HTTP_PROXY'|'AWS_PROXY'
_DEFAULT_TYPE = "AWS_PROXY"
_DEFAULT_INTEGRATION_METHOD = "POST"
_DEFAULT_REQUEST_PARAMETERS = {"integration.request.header.X-Amz-Invocation-Type":
                               "method.request.header.X-Amz-Invocation-Type"}
# ANY, DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT
_DEFAULT_HTTP_METHOD = "ANY"
# NONE, AWS_IAM, CUSTOM, COGNITO_USER_POOLS
_DEFAULT_AUTHORIZATION_TYPE = "NONE"
_DEFAULT_PATH_PART = "{proxy+}"
_DEFAULT_STAGE_NAME = "scar"


class APIGateway(GenericClient):
    """Manage the calls to the ApiGateway client."""

    def __init__(self, aws_properties) -> None:
        super().__init__(aws_properties)
        # {0}: lambda function region, {1}: aws account id, {1}: lambda function name
        self.lambda_uri = "arn:aws:lambda:{region}:{acc_id}:function:{lambdaf_name}/invocations"
        # {0}: api_region, {1}: lambda_uri
        self.uri = "arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/{lambda_uri}"
        # {0}: api_id, {1}: api_region
        self.endpoint = "https://{api_id}.execute-api.{region}.amazonaws.com/scar/launch"

    def _get_uri(self) -> str:
        lambda_uri_kwargs = {'region': self.aws.region,
                             'acc_id': self.aws.account_id,
                             'lambdaf_name': self.aws.lambdaf.name}
        uri_kwargs = {'region': self.aws.region,
                      'lambda_uri': self.lambda_uri.format(**lambda_uri_kwargs)}
        return self.uri.format(**uri_kwargs)

    def _get_common_args(self, resource_info: Dict) -> Dict:
        return {'restApiId' : self.aws.api_gateway.id,
                'resourceId' : resource_info.get('id', ''),
                'httpMethod' : _DEFAULT_HTTP_METHOD}

    def _get_method_args(self, resource_info: Dict) -> Dict:
        args = {'authorizationType' : _DEFAULT_AUTHORIZATION_TYPE,
                'requestParameters' : {'method.request.header.X-Amz-Invocation-Type' : False}}
        method_args = self._get_common_args(resource_info)
        method_args.update(args)
        return method_args

    def _get_integration_args(self, resource_info: Dict) -> Dict:
        args = {'type' : _DEFAULT_TYPE,
                'integrationHttpMethod' : _DEFAULT_INTEGRATION_METHOD,
                'uri' : self._get_uri(),
                'requestParameters' : _DEFAULT_REQUEST_PARAMETERS}
        integration_args = self._get_common_args(resource_info)
        integration_args.update(args)
        return integration_args

    def _get_resource_id(self) -> str:
        res_id = ""
        resources_info = self.client.get_resources(self.aws.api_gateway.id)
        for resource in resources_info['items']:
            if resource['path'] == '/':
                res_id = resource['id']
                break
        return res_id

    def _set_api_gateway_id(self, api_info: Dict) -> None:
        self.aws.api_gateway.id = api_info.get('id', '')

    def _get_endpoint(self) -> str:
        kwargs = {'api_id': self.aws.api_gateway.id, 'region': self.aws.region}
        return self.endpoint.format(**kwargs)

    def create_api_gateway(self) -> None:
        """Creates an Api Gateway endpoint."""
        api_info = self.client.create_rest_api(self.aws.api_gateway.name)
        self._set_api_gateway_id(api_info)
        resource_info = self.client.create_resource(self.aws.api_gateway.id,
                                                    self._get_resource_id(),
                                                    _DEFAULT_PATH_PART)
        self.client.create_method(**self._get_method_args(resource_info))
        self.client.set_integration(**self._get_integration_args(resource_info))
        self.client.create_deployment(self.aws.api_gateway.id, _DEFAULT_STAGE_NAME)
        logger.info(f'API Gateway endpoint: {self._get_endpoint()}')

    def delete_api_gateway(self, api_gateway_id: str) -> None:
        """Deletes an Api Gateway endpoint."""
        return self.client.delete_rest_api(api_gateway_id)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import random
from multiprocessing.pool import ThreadPool
from botocore.exceptions import ClientError
from scar.providers.aws import GenericClient
from scar.providers.aws.functioncode import FunctionPackager
from scar.providers.aws.lambdalayers import LambdaLayers
from scar.providers.aws.s3 import S3
from scar.providers.aws.validators import AWSValidator
import scar.exceptions as excp
import scar.http.request as request
import scar.logger as logger
import scar.providers.aws.response as response_parser
from scar.utils import lazy_property, DataTypesUtils, FileUtils, StrUtils

MAX_CONCURRENT_INVOCATIONS = 500


class Lambda(GenericClient):

    @lazy_property
    def layers(self):
        layers = LambdaLayers(self.client, self.supervisor_version)
        return layers

    @lazy_property
    def s3(self):
        s3 = S3(self.aws)
        return s3

    def __init__(self, aws_properties, supervisor_version):
        super().__init__(aws_properties)
        self.supervisor_version = supervisor_version
        self._initialize_properties(aws_properties)

    def _initialize_properties(self, aws_properties):
        self.aws.lambdaf.environment = {'Variables' : {}}
        self.aws.lambdaf.zip_file_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), 'function.zip')
        self.aws.lambdaf.invocation_type = "RequestResponse"
        self.aws.lambdaf.log_type = "Tail"
        self.aws.lambdaf.layers = []
        if hasattr(self.aws.lambdaf, "name"):
            self.aws.lambdaf.handler = "{0}.lambda_handler".format(self.aws.lambdaf.name)
        if not hasattr(self.aws.lambdaf, "asynchronous"):
            self.aws.lambdaf.asynchronous = False
        self._set_default_call_parameters()

    def _set_default_call_parameters(self):
        self.asynchronous_call_parameters = {"invocation_type" : "Event",
                                             "log_type" : "None",
                                             "asynchronous" : "True"}
        self.request_response_call_parameters = {"invocation_type" : "RequestResponse",
                                                 "log_type" : "Tail",
                                                 "asynchronous" : "False"}

    def _get_creations_args(self):
        return {'FunctionName' : self.aws.lambdaf.name,
                'Runtime' : self.aws.lambdaf.runtime,
                'Role' : self.aws.iam.role,
                'Handler' :  self.aws.lambdaf.handler,
                'Code' : self.aws.lambdaf.code,
                'Environment' : self.aws.lambdaf.environment,
                'Description': self.aws.lambdaf.description,
                'Timeout':  self.aws.lambdaf.time,
                'MemorySize': self.aws.lambdaf.memory,
                'Tags': self.aws.tags,
                'Layers': self.aws.lambdaf.layers}

    def is_asynchronous(self):
        return self.aws.lambdaf.asynchronous

    @excp.exception(logger)
    def create_function(self):
        self._manage_supervisor_layer()
        self._set_environment_variables()
        self._set_function_code()
        creation_args = self._get_creations_args()
        response = self.client.create_function(**creation_args)
        if response and "FunctionArn" in response:
            self.aws.lambdaf.arn = response.get('FunctionArn', "")
        return response

    def _manage_supervisor_layer(self):
        self.layers.check_faas_supervisor_layer()
        self.aws.lambdaf.layers.append(self.layers.get_latest_supervisor_layer_arn())

    def _add_lambda_environment_variable(self, key, value):
        if key and value:
            self.aws.lambdaf.environment['Variables'][key] = value

    def _add_custom_environment_variables(self, env_vars, prefix=''):
            if type(env_vars) is dict:
                for key, val in env_vars.items():
                    # Add an specific prefix to be able to find the variables defined by the user
                    self._add_lambda_environment_variable('{0}{1}'.format(prefix, key), val)
            else:
                for env_var in env_vars:
                    key_val = env_var.split("=")
                    # Add an specific prefix to be able to find the variables defined by the user
                    self._add_lambda_environment_variable('{0}{1}'.format(prefix, key_val[0]), key_val[1])

    def _set_environment_variables(self):
        # Add required variables
        self._set_required_environment_variables()
        # Add explicitly user defined variables
        if hasattr(self.aws.lambdaf, "environment_variables"):
            self._add_custom_environment_variables(self.aws.lambdaf.environment_variables, prefix='CONT_VAR_')
        # Add explicitly user defined variables
        if hasattr(self.aws.lambdaf, "lambda_environment"):
            self._add_custom_environment_variables(self.aws.lambdaf.lambda_environment)

    def _set_required_environment_variables(self):
        self._add_lambda_environment_variable('SUPERVISOR_TYPE', 'LAMBDA')
        self._add_lambda_environment_variable('TIMEOUT_THRESHOLD', str(self.aws.lambdaf.timeout_threshold))
        self._add_lambda_environment_variable('LOG_LEVEL', self.aws.lambdaf.log_level)
        self._add_udocker_variables()
        self._add_execution_mode()
        self._add_s3_environment_vars()
        if hasattr(self.aws.lambdaf, "image"):
            self._add_lambda_environment_variable('IMAGE_ID', self.aws.lambdaf.image)
        if hasattr(self.aws, "api_gateway"):
            self._add_lambda_environment_variable('API_GATEWAY_ID', self.aws.api_gateway.id)

    def _add_udocker_variables(self):
        self._add_lambda_environment_variable('UDOCKER_EXEC', "/opt/udocker/udocker.py")
        self._add_lambda_environment_variable('UDOCKER_DIR', "/tmp/shared/udocker")
        self._add_lambda_environment_variable('UDOCKER_LIB', "/opt/udocker/lib/")
        self._add_lambda_environment_variable('UDOCKER_BIN', "/opt/udocker/bin/")

    def _add_execution_mode(self):
        self._add_lambda_environment_variable('EXECUTION_MODE', self.aws.execution_mode)
        # if (self.aws.execution_mode == 'lambda-batch' or self.aws.execution_mode == 'batch'):
        #     self._add_lambda_environment_variable('BATCH_SUPERVISOR_IMG', self.aws.batch.supervisor_image)

    def _add_s3_environment_vars(self):
        if hasattr(self.aws, "s3"):
            provider_id = random.randint(1, 1000001)

            if hasattr(self.aws.s3, "input_bucket"):
                self._add_lambda_environment_variable(
                    f'STORAGE_PATH_INPUT_{provider_id}',
                    self.aws.s3.storage_path_input
                )

            if hasattr(self.aws.s3, "output_bucket"):
                self._add_lambda_environment_variable(
                    f'STORAGE_PATH_OUTPUT_{provider_id}',
                    self.aws.s3.storage_path_output
                )
            else:
                self._add_lambda_environment_variable(
                    f'STORAGE_PATH_OUTPUT_{provider_id}',
                    self.aws.s3.storage_path_input
                )
            self._add_lambda_environment_variable(
                f'STORAGE_AUTH_S3_USER_{provider_id}',
                'scar'
            )

    @excp.exception(logger)
    def _set_function_code(self):
        # Zip all the files and folders needed
        FunctionPackager(self.aws, self.supervisor_version).create_zip()
        if hasattr(self.aws, "s3") and hasattr(self.aws.s3, 'deployment_bucket'):
            self._upload_to_S3()
            self.aws.lambdaf.code = {"S3Bucket": self.aws.s3.deployment_bucket, "S3Key" : self.aws.s3.file_key}
        else:
            self.aws.lambdaf.code = {"ZipFile": FileUtils.read_file(self.aws.lambdaf.zip_file_path, mode="rb")}

    def _upload_to_S3(self):
        self.aws.s3.input_bucket = self.aws.s3.deployment_bucket
        self.aws.s3.file_key = 'lambda/{0}.zip'.format(self.aws.lambdaf.name)
        self.s3.upload_file(file_path=self.aws.lambdaf.zip_file_path, file_key=self.aws.s3.file_key)

    def delete_function(self, function_name):
        return self.client.delete_function(function_name)

    def link_function_and_input_bucket(self):
        kwargs = {'FunctionName' : self.aws.lambdaf.name,
                  'Principal' : "s3.amazonaws.com",
                  'SourceArn' : 'arn:aws:s3:::{0}'.format(self.aws.s3.input_bucket)}
        self.client.add_invocation_permission(**kwargs)

    def preheat_function(self):
        logger.info("Preheating function")
        self._set_request_response_call_parameters()
        return self.launch_lambda_instance()

    def _launch_async_event(self, s3_event):
        self.set_asynchronous_call_parameters()
        return self._launch_s3_event(s3_event)

    def launch_request_response_event(self, s3_event):
        self._set_request_response_call_parameters()
        return self._launch_s3_event(s3_event)

    def _launch_s3_event(self, s3_event):
        self.aws.lambdaf.payload = s3_event
        logger.info(f"Sending event for file '{s3_event['Records'][0]['s3']['object']['key']}'")
        return self.launch_lambda_instance()

    def process_asynchronous_lambda_invocations(self, s3_event_list):
        if (len(s3_event_list) > MAX_CONCURRENT_INVOCATIONS):
            s3_file_chunk_list = DataTypesUtils.divide_list_in_chunks(s3_event_list, MAX_CONCURRENT_INVOCATIONS)
            for s3_file_chunk in s3_file_chunk_list:
                self._launch_concurrent_lambda_invocations(s3_file_chunk)
        else:
            self._launch_concurrent_lambda_invocations(s3_event_list)

    def _launch_concurrent_lambda_invocations(self, s3_event_list):
        pool = ThreadPool(processes=len(s3_event_list))
        pool.map(lambda s3_event: self._launch_async_event(s3_event), s3_event_list)
        pool.close()

    def launch_lambda_instance(self):
        response = self._invoke_lambda_function()
        response_args = {'Response' : response,
                         'FunctionName' : self.aws.lambdaf.name,
                         'OutputType' : self.aws.output,
                         'IsAsynchronous' : self.aws.lambdaf.asynchronous}
        if hasattr(self.aws, "output_file"):
            response_args['OutputFile'] = self.aws.output_file
        response_parser.parse_invocation_response(**response_args)

    def _get_invocation_payload(self):
        # Default payload
        payload = self.aws.lambdaf.payload if hasattr(self.aws.lambdaf, 'payload') else {}
        if not payload:            
            # Check for defined run script
            if hasattr(self.aws.lambdaf, "run_script"):
                script_path = self.aws.lambdaf.run_script
                if hasattr(self.aws, "config_path"):
                    script_path = FileUtils.join_paths(self.aws.config_path, script_path)
                # We first code to base64 in bytes and then decode those bytes to allow the json lib to parse the data
                # https://stackoverflow.com/questions/37225035/serialize-in-json-a-base64-encoded-data#37239382
                payload = { "script" : StrUtils.bytes_to_base64str(FileUtils.read_file(script_path, 'rb')) }
            # Check for defined commands
            # This overrides any other function payload
            if hasattr(self.aws.lambdaf, "c_args"):
                payload = { "cmd_args" : json.dumps(self.aws.lambdaf.c_args) }
        return json.dumps(payload)

    def _invoke_lambda_function(self):
        invoke_args = {'FunctionName' :  self.aws.lambdaf.name,
                       'InvocationType' :  self.aws.lambdaf.invocation_type,
                       'LogType' :  self.aws.lambdaf.log_type,
                       'Payload' : self._get_invocation_payload()}
        return self.client.invoke_function(**invoke_args)

    def set_asynchronous_call_parameters(self):
        self.aws.lambdaf.update_properties(**self.asynchronous_call_parameters)

    def _set_request_response_call_parameters(self):
        self.aws.lambdaf.update_properties(**self.request_response_call_parameters)

    def _update_environment_variables(self, function_info, update_args):
        # To update the environment variables we need to retrieve the
        # variables defined in lambda and update them with the new values
        env_vars = self.aws.lambdaf.environment
        if hasattr(self.aws.lambdaf, "environment_variables"):
            for env_var in self.aws.lambdaf.environment_variables:
                key_val = env_var.split("=")
                # Add an specific prefix to be able to find the variables defined by the user
                env_vars['Variables']['CONT_VAR_{0}'.format(key_val[0])] = key_val[1]
        if hasattr(self.aws.lambdaf, "timeout_threshold"):
            env_vars['Variables']['TIMEOUT_THRESHOLD'] = str(self.aws.lambdaf.timeout_threshold)
        if hasattr(self.aws.lambdaf, "log_level"):
            env_vars['Variables']['LOG_LEVEL'] = self.aws.lambdaf.log_level
        function_info['Environment']['Variables'].update(env_vars['Variables'])
        update_args['Environment'] = function_info['Environment']

    def _update_supervisor_layer(self, function_info, update_args):
        if hasattr(self.aws.lambdaf, "supervisor_layer"):
            # Set supervisor layer Arn
            function_layers = [self.layers.get_latest_supervisor_layer_arn()]
            # Add the rest of layers (if exist)
            if 'Layers' in function_info:
                function_layers.extend([layer for layer in function_info['Layers'] if self.layers.layer_name not in layer['Arn']])
            update_args['Layers'] = function_layers

    def update_function_configuration(self, function_info=None):
        if not function_info:
            function_info = self.get_function_info()
        update_args = {'FunctionName' : function_info['FunctionName'] }
#         if hasattr(self.aws.lambdaf, "memory"):
#             update_args['MemorySize'] = self.aws.lambdaf.memory
#         else:
#             update_args['MemorySize'] = function_info['MemorySize']
#         if hasattr(self.aws.lambdaf, "time"):
#             update_args['Timeout'] = self.aws.lambdaf.time
#         else:
#             update_args['Timeout'] = function_info['Timeout']
        self._update_environment_variables(function_info, update_args)
        self._update_supervisor_layer(function_info, update_args)
        self.client.update_function_configuration(**update_args)
        logger.info("Function '{}' updated successfully.".format(function_info['FunctionName']))

    def _get_function_environment_variables(self):
        return self.get_function_info()['Environment']

    def get_all_functions(self, arn_list):
        try:
            return [self.get_function_info(function_arn) for function_arn in arn_list]
        except ClientError as cerr:
            print (f"Error getting function info by arn: {cerr}")

    def get_function_info(self, function_name_or_arn=None):
        name_arn = function_name_or_arn if function_name_or_arn else self.aws.lambdaf.name
        return self.client.get_function_info(name_arn)

    @excp.exception(logger)
    def find_function(self, function_name_or_arn=None):
        try:
            # If this call works the function exists
            name_arn = function_name_or_arn if function_name_or_arn else self.aws.lambdaf.name
            self.get_function_info(name_arn)
            return True
        except ClientError as ce:
            # Function not found
            if ce.response['Error']['Code'] == 'ResourceNotFoundException':
                return False
            else:
                raise

    def add_invocation_permission_from_api_gateway(self):
        kwargs = {'FunctionName' : self.aws.lambdaf.name,
                  'Principal' : 'apigateway.amazonaws.com',
                  'SourceArn' : 'arn:aws:execute-api:{0}:{1}:{2}/*'.format(self.aws.region,
                                                                           self.aws.account_id,
                                                                           self.aws.api_gateway.id)}
        # Add Testing permission
        self.client.add_invocation_permission(**kwargs)
        # Add Invocation permission
        kwargs['SourceArn'] = 'arn:aws:execute-api:{0}:{1}:{2}/scar/ANY'.format(self.aws.region,
                                                                                self.aws.account_id,
                                                                                self.aws.api_gateway.id)
        self.client.add_invocation_permission(**kwargs)

    def get_api_gateway_id(self):
        env_vars = self._get_function_environment_variables()
        return env_vars['Variables'].get('API_GATEWAY_ID', '')

    def _get_api_gateway_url(self):
        api_id = self.get_api_gateway_id()
        if not api_id:
            raise excp.ApiEndpointNotFoundError(self.aws.lambdaf.name)
        return f'https://{api_id}.execute-api.{self.aws.region}.amazonaws.com/scar/launch'

    def call_http_endpoint(self):
        invoke_args = {'headers' : {'X-Amz-Invocation-Type':'Event'} if self.is_asynchronous() else {}}
        if hasattr(self.aws, "api_gateway"):
            self._set_invoke_args(invoke_args)
        return request.call_http_endpoint(self._get_api_gateway_url(), **invoke_args)

    def _set_invoke_args(self, invoke_args):
        if hasattr(self.aws.api_gateway, "data_binary"):
            invoke_args['data'] = self._get_b64encoded_binary_data(self.aws.api_gateway.data_binary)
            invoke_args['headers'] = {'Content-Type': 'application/octet-stream'}
        if hasattr(self.aws.api_gateway, "parameters"):
            invoke_args['params'] = self._parse_http_parameters(self.aws.api_gateway.parameters)
        if hasattr(self.aws.api_gateway, "json_data"):
            invoke_args['data'] = self._parse_http_parameters(self.aws.api_gateway.json_data)
            invoke_args['headers'] = {'Content-Type': 'application/json'}

    def _parse_http_parameters(self, parameters):
        return parameters if type(parameters) is dict else json.loads(parameters)

    @excp.exception(logger)
    def _get_b64encoded_binary_data(self, data_path):
        if data_path:
            AWSValidator.validate_http_payload_size(data_path, self.is_asynchronous())
            with open(data_path, 'rb') as data_file:
                return base64.b64encode(data_file.read())



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import Dict, List
from scar.providers.aws import GenericClient
import scar.logger as logger
from scar.providers.aws.launchtemplates import LaunchTemplates
from scar.utils import lazy_property, FileUtils, StrUtils

_LAUNCH_TEMPLATE_NAME = 'faas-supervisor'


def _get_job_definitions(jobs_info: Dict) -> List:
    return [f"{job_def.get('jobDefinitionName', '')}:{job_def.get('revision', '')}"
            for job_def in jobs_info.get('jobDefinitions', {})]


class Batch(GenericClient):

    @lazy_property
    def launch_templates(self):
        launch_templates = LaunchTemplates(self.aws, self.supervisor_version)
        return launch_templates

    def __init__(self, aws_properties, supervisor_version):
        super().__init__(aws_properties)
        self.supervisor_version = supervisor_version
        self.aws.batch.instance_role = (f"arn:aws:iam::{self.aws.account_id}:"
                                        "instance-profile/ecsInstanceRole")
        self.aws.batch.service_role = (f"arn:aws:iam::{self.aws.account_id}:"
                                       "role/service-role/AWSBatchServiceRole")
        self.aws.batch.env_vars = []

    def _set_required_environment_variables(self):
        self._set_batch_environment_variable('AWS_LAMBDA_FUNCTION_NAME', self.aws.lambdaf.name)
        self._set_batch_environment_variable('SCRIPT', self._get_user_script())
        if (hasattr(self.aws.lambdaf, 'environment_variables') and
                self.aws.lambdaf.environment_variables):
            self._add_custom_environment_variables(self.aws.lambdaf.environment_variables)
        if (hasattr(self.aws.lambdaf, 'lambda_environment') and
                self.aws.lambdaf.lambda_environment):
            self._add_custom_environment_variables(self.aws.lambdaf.lambda_environment)
        if hasattr(self.aws, "s3"):
            self._add_s3_environment_vars()

    def _add_custom_environment_variables(self, env_vars):
        if isinstance(env_vars, dict):
            for key, val in env_vars.items():
                self._set_batch_environment_variable(key, val)
        else:
            for env_var in env_vars:
                self._set_batch_environment_variable(*env_var.split("="))

    def _set_batch_environment_variable(self, key, value):
        if key and value is not None:
            self.aws.batch.env_vars.append({'name': key, 'value': value})

    def _add_s3_environment_vars(self):
        provider_id = random.randint(1, 1000001)
        if hasattr(self.aws.s3, "input_bucket"):
            self._set_batch_environment_variable(f'STORAGE_PATH_INPUT_{provider_id}',
                                                 self.aws.s3.storage_path_input)
        if hasattr(self.aws.s3, "output_bucket"):
            self._set_batch_environment_variable(f'STORAGE_PATH_OUTPUT_{provider_id}',
                                                 self.aws.s3.storage_path_output)
        else:
            self._set_batch_environment_variable(f'STORAGE_PATH_OUTPUT_{provider_id}',
                                                 self.aws.s3.storage_path_input)
        self._set_batch_environment_variable(f'STORAGE_AUTH_S3_USER_{provider_id}', 'scar')

    def _get_user_script(self):
        script = ''
        if hasattr(self.aws.lambdaf, "init_script"):
            file_content = FileUtils.read_file(self.aws.lambdaf.init_script)
            script = StrUtils.utf8_to_base64_string(file_content)
        return script

    def _delete_job_definitions(self, name):
        job_definitions = []
        # Get IO definitions (if any)
        kwargs = {"jobDefinitionName": '{0}-io'.format(name)}
        io_job_info = self.client.describe_job_definitions(**kwargs)
        job_definitions.extend(_get_job_definitions(io_job_info))
        # Get main job definition
        kwargs = {"jobDefinitionName": name}
        job_info = self.client.describe_job_definitions(**kwargs)
        job_definitions.extend(_get_job_definitions(job_info))
        for job_def in job_definitions:
            kwars = {"jobDefinition": job_def}
            self.client.deregister_job_definition(**kwars)
        logger.info("Job definitions deleted")

    def _get_job_queue_info(self, name):
        job_queue_info_args = {'jobQueues': [self._get_resource_name(name)]}
        return self.client.describe_job_queues(**job_queue_info_args)

    def _delete_job_queue(self, name):
        response = self._get_job_queue_info(name)
        while response["jobQueues"]:
            state = response["jobQueues"][0]["state"]
            status = response["jobQueues"][0]["status"]
            if status == "VALID":
                self._delete_valid_job_queue(state, name)
            response = self._get_job_queue_info(name)

    def _delete_valid_job_queue(self, state, name):
        if state == "ENABLED":
            updating_args = {'jobQueue': self._get_resource_name(name),
                             'state': 'DISABLED'}
            self.client.update_job_queue(**updating_args)
        elif state == "DISABLED":
            deleting_args = {'jobQueue': self._get_resource_name(name)}
            logger.info("Job queue deleted")
            self.client.delete_job_queue(**deleting_args)

    def _get_compute_env_info(self, name):
        creation_args = self._get_describe_compute_env_args(name_c=name)
        return self.client.describe_compute_environments(**creation_args)

    def _delete_compute_env(self, name):
        response = self._get_compute_env_info(name)
        while response["computeEnvironments"]:
            state = response["computeEnvironments"][0]["state"]
            status = response["computeEnvironments"][0]["status"]
            if status == "VALID":
                self._delete_valid_compute_environment(state, name)
            response = self._get_compute_env_info(name)

    def _delete_valid_compute_environment(self, state, name):
        if state == "ENABLED":
            update_args = {'computeEnvironment': self._get_resource_name(name),
                           'state': 'DISABLED'}
            self.client.update_compute_environment(**update_args)
        elif state == "DISABLED":
            delete_args = {'computeEnvironment': self._get_resource_name(name)}
            logger.info("Compute environment deleted")
            self.client.delete_compute_environment(**delete_args)

    def _get_compute_env_args(self):
        return {
            'computeEnvironmentName': self.aws.lambdaf.name,
            'serviceRole': self.aws.batch.service_role,
            'type': self.aws.batch.compute_resources['type'],
            'state':  self.aws.batch.compute_resources['state'],
            'computeResources': {
                'type': self.aws.batch.compute_resources['comp_type'],
                'minvCpus': self.aws.batch.compute_resources['min_v_cpus'],
                'maxvCpus': self.aws.batch.compute_resources['max_v_cpus'],
                'desiredvCpus': self.aws.batch.compute_resources['desired_v_cpus'],
                'instanceTypes': self.aws.batch.compute_resources['instance_types'],
                'subnets': self.aws.batch.compute_resources['subnets'],
                'securityGroupIds': self.aws.batch.compute_resources['security_group_ids'],
                'instanceRole': self.aws.batch.instance_role,
                'launchTemplate': {
                    'launchTemplateName': _LAUNCH_TEMPLATE_NAME,
                    'version': str(self.launch_templates.get_launch_template_version())
                }
            }
        }

    def _get_creations_job_queue_args(self):
        return {
            'computeEnvironmentOrder': [{'computeEnvironment': self.aws.lambdaf.name,
                                         'order': 1}, ],
            'jobQueueName':  self.aws.lambdaf.name,
            'priority': 1,
            'state': self.aws.batch.compute_resources['state'],
        }

    def _get_resource_name(self, name=None):
        return name if name else self.aws.lambdaf.name

    def _get_describe_compute_env_args(self, name_c=None):
        return {'computeEnvironments': [self._get_resource_name(name_c)]}

    def _get_job_definition_args(self):
        job_def_args = {
            'jobDefinitionName': self.aws.lambdaf.name,
            'type': 'container',
            'containerProperties': {
                'image': self.aws.lambdaf.image,
                'memory': int(self.aws.batch.memory),
                'vcpus': int(self.aws.batch.vcpus),
                'command': [
                    '/bin/sh',
                    '-c',
                    'echo $EVENT | /opt/faas-supervisor/bin/supervisor'
                ],
                'volumes': [
                    {
                        'host': {
                            'sourcePath': '/opt/faas-supervisor/bin'
                        },
                        'name': 'supervisor-bin'
                    }
                ],
                'environment': self.aws.batch.env_vars,
                'mountPoints': [
                    {
                        'containerPath': '/opt/faas-supervisor/bin',
                        'sourceVolume': 'supervisor-bin'
                    }
                ]
            }
        }
        if self.aws.batch.enable_gpu:
            job_def_args['containerProperties']['resourceRequirements'] = [
                {
                    'value': '1',
                    'type': 'GPU'
                }
            ]
        return job_def_args

    def _get_state_and_status_of_compute_env(self, name=None):
        creation_args = self._get_describe_compute_env_args(name_c=name)
        response = self.client.describe_compute_environments(**creation_args)
        return (response["computeEnvironments"][0]["state"],
                response["computeEnvironments"][0]["status"])

    def create_batch_environment(self):
        self._set_required_environment_variables()
        creation_args = self._get_compute_env_args()
        self.client.create_compute_environment(**creation_args)
        while True:
            state, status = self._get_state_and_status_of_compute_env()
            if state == "ENABLED" and status == "VALID":
                creation_args = self._get_creations_job_queue_args()
                logger.info("Compute environment successfully created.")
                self.client.create_job_queue(**creation_args)
                logger.info('Job queue successfully created.')
                creation_args = self._get_job_definition_args()
                logger.info(f"Registering '{self.aws.lambdaf.name}' job definition.")
                return self.client.register_job_definition(**creation_args)

    def delete_compute_environment(self, name):
        self._delete_job_definitions(name)
        self._delete_job_queue(name)
        self._delete_compute_env(name)

    def exist_compute_environments(self, name):
        creation_args = self._get_describe_compute_env_args(name_c=name)
        response = self.client.describe_compute_environments(**creation_args)
        return len(response["computeEnvironments"]) > 0

    def describe_jobs(self, job_id):
        describe_args = {'jobs': [job_id]}
        return self.client.describe_jobs(**describe_args)

    def exist_job(self, job_id):
        response = self.describe_jobs(job_id)
        return len(response["jobs"]) != 0



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes in charge
of managing the SCAR configuration file."""

import json
from scar.exceptions import exception, ScarConfigFileError
import scar.logger as logger
from scar.utils import FileUtils, SysUtils, StrUtils

_DEFAULT_CFG = {
    "scar": {
        # Must be a tag or "latest"
        "supervisor_version": "latest",
        "config_version": "1.0.0"
    },
    "aws": {
        "boto_profile": "default",
        "region": "us-east-1",
        "execution_mode": "lambda",
        "iam": {"role": ""},
        "lambda": {
            "time": 300,
            "memory": 512,
            "description": "Automatically generated lambda function",
            "timeout_threshold": 10,
            "runtime": "python3.6",
            "max_payload_size": 52428800,
            "max_s3_payload_size": 262144000,
            "layers": []
        },
        "cloudwatch": {"log_retention_policy_in_days": 30},
        "batch": {
            "vcpus": 1,
            "memory": 1024,
            "enable_gpu": False,
            "compute_resources": {
                "state": "ENABLED",
                "type": "MANAGED",
                "security_group_ids": [],
                "comp_type": "EC2",
                "desired_v_cpus": 0,
                "min_v_cpus": 0,
                "max_v_cpus": 2,
                "subnets": [],
                "instance_types": ["m3.medium"]
            }
        }
    }
}


class ConfigFileParser():
    """Class to manage the SCAR configuration file creation, update and load."""

    _CONFIG_FOLDER_PATH = ".scar"
    _CONFIG_FILE_PATH = "scar.cfg"
    _CONFIG_FILE_NAME_BCK = "scar.cfg_old"
    config_file_folder = FileUtils.join_paths(SysUtils.get_user_home_path(), _CONFIG_FOLDER_PATH)
    config_file_path = FileUtils.join_paths(config_file_folder, _CONFIG_FILE_PATH)
    backup_file_path = FileUtils.join_paths(config_file_folder, _CONFIG_FILE_NAME_BCK)

    @exception(logger)
    def __init__(self):
        # Check if the config file exists
        if FileUtils.is_file(self.config_file_path):
            with open(self.config_file_path) as cfg_file:
                self.cfg_data = json.load(cfg_file)
            if not self._is_config_file_updated():
                self._update_config_file()
        else:
            self._create_scar_config_folder_and_file()

    def _is_config_file_updated(self):
        if 'config_version' not in self.cfg_data['scar']:
            return False
        return StrUtils.compare_versions(self.cfg_data.get('scar', {}).get("config_version", ""),
                                         _DEFAULT_CFG['scar']["config_version"]) <= 0

    def get_properties(self):
        """Returns the configuration data of the configuration file."""
        return self.cfg_data

    def get_udocker_zip_url(self):
        """Returns the url where the udocker zip is stored."""
        return self.cfg_data['scar']['udocker_info']['zip_url']

    def _create_scar_config_folder_and_file(self):
        FileUtils.create_folder(self.config_file_folder)
        self._create_new_config_file()
        raise ScarConfigFileError(file_path=self.config_file_path)

    def _create_new_config_file(self):
        FileUtils.create_file_with_content(self.config_file_path,
                                           json.dumps(_DEFAULT_CFG, indent=2))

    def _update_config_file(self):
        logger.info(("SCAR configuration file deprecated.\n"
                     "Updating your SCAR configuration file."))
        FileUtils.copy_file(self.config_file_path, self.backup_file_path)
        logger.info(f"Old configuration file saved in '{self.backup_file_path}'.")
        self._create_new_config_file()
        logger.info((f"New configuration file saved in '{self.config_file_path}'.\n"
                     "Please fill your new configuration file with your account information."))
        SysUtils.finish_scar_execution()

#         self._merge_files(self.cfg_data, _DEFAULT_CFG)
#         self._delete_unused_data()
#         with open(self.config_file_path, mode='w') as cfg_file:
#             cfg_file.write(json.dumps(self.cfg_data, indent=2))

#     def _add_missing_attributes(self):
#         logger.info("Updating old scar config file '{0}'.\n".format(self.config_file_path))
#         FileUtils.copy_file(self.config_file_path, self.backup_file_path)
#         logger.info("Old scar config file saved in '{0}'.\n".format(self.backup_file_path))
#         self._merge_files(self.cfg_data, _DEFAULT_CFG)
#         self._delete_unused_data()
#         with open(self.config_file_path, mode='w') as cfg_file:
#             cfg_file.write(json.dumps(self.cfg_data, indent=2))
#
#     def _merge_files(self, cfg_data, default_data):
#         for key, val in default_data.items():
#             if key not in cfg_data:
#                 cfg_data[key] = val
#             elif isinstance(cfg_data[key], dict):
#                 self._merge_files(cfg_data[key], default_data[key])
#
#     def _delete_unused_data(self):
#         if 'region' in self.cfg_data['aws']['lambda']:
#             region = self.cfg_data['aws']['lambda'].pop('region', None)
#             if region:
#                 self.cfg_data['aws']['region'] = region



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zipfile import ZipFile
from scar.utils import FileUtils, SysUtils


def _extract_udocker_zip(supervisor_zip_path) -> None:
    file_path = ""
    with ZipFile(supervisor_zip_path) as thezip:
        for file in thezip.namelist():
            if file.endswith("udocker.zip"):
                file_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), file)
                thezip.extract(file, FileUtils.get_tmp_dir())
                break
    return file_path


class Udocker():

    def __init__(self, aws_properties, function_tmp_folder, supervisor_zip_path):
        self.aws = aws_properties
        self.function_tmp_folder = function_tmp_folder
        self.udocker_dir = FileUtils.join_paths(self.function_tmp_folder, "udocker")
        self.udocker_dir_orig = ""
        self._initialize_udocker(supervisor_zip_path)

    def _initialize_udocker(self, supervisor_zip_path):
        self.udocker_code = FileUtils.join_paths(self.udocker_dir, "udocker.py")
        self.udocker_exec = ['python3', self.udocker_code]
        self._install_udocker(supervisor_zip_path)

    def _install_udocker(self, supervisor_zip_path):
        udocker_zip_path = _extract_udocker_zip(supervisor_zip_path)
        with ZipFile(udocker_zip_path) as thezip:
            thezip.extractall(self.function_tmp_folder)

    def save_tmp_udocker_env(self):
        # Avoid override global variables
        if SysUtils.is_variable_in_environment("UDOCKER_DIR"):
            self.udocker_dir_orig = SysUtils.get_environment_variable("UDOCKER_DIR")
        # Set temporal global vars
        SysUtils.set_environment_variable("UDOCKER_DIR", self.udocker_dir)

    def restore_udocker_env(self):
        if self.udocker_dir_orig:
            SysUtils.set_environment_variable("UDOCKER_DIR", self.udocker_dir_orig)
        else:
            SysUtils.delete_environment_variable("UDOCKER_DIR")

    def _set_udocker_local_registry(self):
        self.aws.lambdaf.environment['Variables']['UDOCKER_REPOS'] = '/var/task/udocker/repos/'
        self.aws.lambdaf.environment['Variables']['UDOCKER_LAYERS'] = '/var/task/udocker/layers/'

    def _create_udocker_container(self):
        """Check if the container fits in the limits of the deployment."""
        if hasattr(self.aws, "s3") and hasattr(self.aws.s3, "deployment_bucket"):
            self._validate_container_size(self.aws.lambdaf.max_s3_payload_size)
        else:
            self._validate_container_size(self.aws.lambdaf.max_payload_size)

    def _validate_container_size(self, max_payload_size):
        if FileUtils.get_tree_size(self.udocker_dir) < (max_payload_size / 2):
            ucmd = self.udocker_exec + ["create", "--name=lambda_cont", self.aws.lambdaf.image]
            SysUtils.execute_command_with_msg(ucmd, cli_msg="Creating container structure")

        elif FileUtils.get_tree_size(self.udocker_dir) > max_payload_size:
            FileUtils.delete_folder(FileUtils.join_paths(self.udocker_dir, "containers"))

        else:
            self.aws.lambdaf.environment['Variables']['UDOCKER_LAYERS'] = \
                                         '/var/task/udocker/containers/'

    def download_udocker_image(self):
        self.save_tmp_udocker_env()
        SysUtils.execute_command_with_msg(self.udocker_exec + ["pull", self.aws.lambdaf.image],
                                          cli_msg="Downloading container image")
        self._create_udocker_container()
        self._set_udocker_local_registry()
        self.restore_udocker_env()

    def prepare_udocker_image(self):
        self.save_tmp_udocker_env()
        image_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), "udocker_image.tar.gz")
        FileUtils.copy_file(self.aws.lambdaf.image_file, image_path)
        cmd_out = SysUtils.execute_command_with_msg(self.udocker_exec + ["load", "-i", image_path],
                                                    cli_msg="Loading image file")
        # Get the image name from the command output
        self.aws.lambdaf.image = cmd_out.split('\n')[1]
        self._create_udocker_container()
        self.aws.lambdaf.environment['Variables']['IMAGE_ID'] = self.aws.lambdaf.image
        self._set_udocker_local_registry()
        self.restore_udocker_env()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import scar.exceptions as excp
import scar.logger as logger
from scar.utils import DataTypesUtils
import scar.version as version


class CommandParser(object):

    def __init__(self, scar_cli):
        self.scar_cli = scar_cli
        self.create_parser()
        self.create_parent_parsers()
        self.add_subparsers()

    def create_parser(self):
        self.parser = argparse.ArgumentParser(prog="scar",
                                              description="Deploy containers in serverless architectures",
                                              epilog="Run 'scar COMMAND --help' for more information on a command.")
        self.parser.add_argument('--version', help='Show SCAR version.', dest="version", action="store_true", default=False)

    def create_parent_parsers(self):
        self.create_function_definition_parser()
        self.create_exec_parser()
        self.create_output_parser()
        self.create_profile_parser()
        self.create_storage_parser()

    def create_function_definition_parser(self):
        self.function_definition_parser = argparse.ArgumentParser(add_help=False)
        self.function_definition_parser.add_argument("-d", "--description", help="Lambda function description.")
        self.function_definition_parser.add_argument("-e", "--environment", action='append', help="Pass environment variable to the container (VAR=val). Can be defined multiple times.")
        self.function_definition_parser.add_argument("-le", "--lambda-environment", action='append', help="Pass environment variable to the lambda function (VAR=val). Can be defined multiple times.")
        self.function_definition_parser.add_argument("-m", "--memory", type=int, help="Lambda function memory in megabytes. Range from 128 to 3008 in increments of 64")
        self.function_definition_parser.add_argument("-t", "--time", type=int, help="Lambda function maximum execution time in seconds. Max 900.")
        self.function_definition_parser.add_argument("-tt", "--timeout-threshold", type=int, help="Extra time used to postprocess the data. This time is extracted from the total time of the lambda function.")
        self.function_definition_parser.add_argument("-ll", "--log-level", help="Set the log level of the lambda function. Accepted values are: 'CRITICAL','ERROR','WARNING','INFO','DEBUG'", default="INFO")
        self.function_definition_parser.add_argument("-l", "--layers", action='append', help="Pass layers ARNs to the lambda function. Can be defined multiple times.")
        self.function_definition_parser.add_argument("-ib", "--input-bucket", help="Bucket name where the input files will be stored.")
        self.function_definition_parser.add_argument("-ob", "--output-bucket", help="Bucket name where the output files are saved.")
        self.function_definition_parser.add_argument("-em", "--execution-mode", help="Specifies the execution mode of the job. It can be 'lambda', 'lambda-batch' or 'batch'")
        self.function_definition_parser.add_argument("-r", "--iam-role", help="IAM role used in the management of the functions")
        self.function_definition_parser.add_argument("-sv", "--supervisor-version", help="FaaS Supervisor version. Can be a tag or 'latest'.")
        # Batch (job definition) options
        self.function_definition_parser.add_argument("-bm", "--batch-memory", help="Batch job memory in megabytes")
        self.function_definition_parser.add_argument("-bc", "--batch-vcpus", help="Number of vCPUs reserved for the Batch container")
        self.function_definition_parser.add_argument("-g", "--enable-gpu", help="Reserve one physical GPU for the Batch container (if it's available in the compute environment)", action="store_true")

    def create_exec_parser(self):
        self.exec_parser = argparse.ArgumentParser(add_help=False)
        self.exec_parser.add_argument("-a", "--asynchronous", help="Launch an asynchronous function.", action="store_true")
        self.exec_parser.add_argument("-o", "--output-file", help="Save output as a file")

    def create_output_parser(self):
        self.output_parser = argparse.ArgumentParser(add_help=False)
        self.output_parser.add_argument("-j", "--json", help="Return data in JSON format", action="store_true")
        self.output_parser.add_argument("-v", "--verbose", help="Show the complete aws output in json format", action="store_true")

    def create_profile_parser(self):
        self.profile_parser = argparse.ArgumentParser(add_help=False)
        self.profile_parser.add_argument("-pf", "--profile", help="AWS profile to use")

    def create_storage_parser(self):
        self.storage_parser = argparse.ArgumentParser(add_help=False)
        self.storage_parser.add_argument("-b", "--bucket", help="Bucket to use as storage", required=True)
        self.storage_parser.add_argument("-p", "--path", help="Path of the file or folder", required=True)

    def add_subparsers(self):
        self.subparsers = self.parser.add_subparsers(title='Commands')
        self.add_init_parser()
        self.add_invoke_parser()
        self.add_run_parser()
        self.add_update_parser()
        self.add_rm_parser()
        self.add_ls_parser()
        self.add_log_parser()
        self.add_put_parser()
        self.add_get_parser()

    def add_init_parser(self):
        parser_init = self.subparsers.add_parser('init', parents=[self.function_definition_parser, self.output_parser, self.profile_parser], help="Create lambda function")
        # Set default function
        parser_init.set_defaults(func=self.scar_cli.init)
        # Lambda conf
        group = parser_init.add_mutually_exclusive_group(required=True)
        group.add_argument("-i", "--image", help="Container image id (i.e. centos:7)")
        group.add_argument("-if", "--image-file", help="Container image file created with 'docker save' (i.e. centos.tar.gz)")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")
        parser_init.add_argument("-n", "--name", help="Lambda function name")
        parser_init.add_argument("-s", "--init-script", help="Path to the input file passed to the function")
        parser_init.add_argument("-ph", "--preheat", help="Preheats the function running it once and downloading the necessary container", action="store_true")
        parser_init.add_argument("-ep", "--extra-payload", help="Folder containing files that are going to be added to the lambda function")
        parser_init.add_argument("-db", "--deployment-bucket", help="Bucket where the deployment package is going to be uploaded.")
        # API Gateway conf
        parser_init.add_argument("-api", "--api-gateway-name", help="API Gateway name created to launch the lambda function")

    def add_invoke_parser(self):
        parser_invoke = self.subparsers.add_parser('invoke', parents=[self.profile_parser, self.exec_parser], help="Call a lambda function using an HTTP request")
        # Set default function
        parser_invoke.set_defaults(func=self.scar_cli.invoke)
        group = parser_invoke.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")
        parser_invoke.add_argument("-db", "--data-binary", help="File path of the HTTP data to POST.")
        parser_invoke.add_argument("-jd", "--json-data", help="JSON Body to Post")
        parser_invoke.add_argument("-p", "--parameters", help="In addition to passing the parameters in the URL, you can pass the parameters here (i.e. '{\"key1\": \"value1\", \"key2\": [\"value2\", \"value3\"]}').")

    def add_update_parser(self):
        parser_update = self.subparsers.add_parser('update', parents=[self.function_definition_parser, self.output_parser, self.profile_parser], help="Update function properties")
        parser_update.set_defaults(func=self.scar_cli.update)
        group = parser_update.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-a", "--all", help="Update all lambda functions", action="store_true")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")

    def add_run_parser(self):
        parser_run = self.subparsers.add_parser('run', parents=[self.output_parser, self.profile_parser, self.exec_parser], help="Deploy function")
        parser_run.set_defaults(func=self.scar_cli.run)
        group = parser_run.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")
        parser_run.add_argument("-s", "--run-script", help="Path to the input file passed to the function")
        parser_run.add_argument('c_args', nargs=argparse.REMAINDER, help="Arguments passed to the container.")

    def add_rm_parser(self):
        parser_rm = self.subparsers.add_parser('rm', parents=[self.output_parser, self.profile_parser], help="Delete function")
        parser_rm.set_defaults(func=self.scar_cli.rm)
        group = parser_rm.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-a", "--all", help="Delete all lambda functions", action="store_true")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")

    def add_log_parser(self):
        parser_log = self.subparsers.add_parser('log', parents=[self.profile_parser], help="Show the logs for the lambda function")
        parser_log.set_defaults(func=self.scar_cli.log)
        group = parser_log.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")
        # CloudWatch args
        parser_log.add_argument("-ls", "--log-stream-name", help="Return the output for the log stream specified.")
        parser_log.add_argument("-ri", "--request-id", help="Return the output for the request id specified.")

    def add_ls_parser(self):
        parser_ls = self.subparsers.add_parser('ls', parents=[self.output_parser, self.profile_parser], help="List lambda functions")
        parser_ls.set_defaults(func=self.scar_cli.ls)
        # S3 args
        parser_ls.add_argument("-b", "--bucket", help="Show bucket files")
        # Layer args
        parser_ls.add_argument("-l", "--list-layers", help="Show lambda layers information", action="store_true")

    def add_put_parser(self):
        parser_put = self.subparsers.add_parser('put', parents=[self.storage_parser, self.profile_parser], help="Upload file(s) to bucket")
        parser_put.set_defaults(func=self.scar_cli.put)

    def add_get_parser(self):
        parser_get = self.subparsers.add_parser('get', parents=[self.storage_parser, self.profile_parser], help="Download file(s) from bucket")
        parser_get.set_defaults(func=self.scar_cli.get)

    @excp.exception(logger)
    def parse_arguments(self):
        """Command parsing and selection"""
        try:
            cmd_args = self.parser.parse_args()
            if cmd_args.version:
                print(f"SCAR {version.__version__}")
                sys.exit(0)

            cmd_args = vars(cmd_args)
            if 'func' not in cmd_args:
                raise excp.MissingCommandError()
            scar_args = self.parse_scar_args(cmd_args)
            aws_args = self.parse_aws_args(cmd_args)
            return DataTypesUtils.merge_dicts(scar_args, aws_args)
        except AttributeError as aerr:
            logger.error("Incorrect arguments: use scar -h to see the options available",
                         f"Error parsing arguments: {aerr}")
        else:
            raise

    def set_args(self, args, key, val):
        if key and val:
            args[key] = val

    def parse_aws_args(self, cmd_args):
        aws_args = {}
        other_args = [('profile', 'boto_profile'), 'region', 'execution_mode']
        self.set_args(aws_args, 'iam', self.parse_iam_args(cmd_args))
        self.set_args(aws_args, 'lambda', self.parse_lambda_args(cmd_args))
        self.set_args(aws_args, 'batch', self.parse_batch_args(cmd_args))
        self.set_args(aws_args, 'cloudwatch', self.parse_cloudwatchlogs_args(cmd_args))
        self.set_args(aws_args, 's3', self.parse_s3_args(cmd_args))
        self.set_args(aws_args, 'api_gateway', self.parse_api_gateway_args(cmd_args))
        aws_args.update(DataTypesUtils.parse_arg_list(other_args, cmd_args))
        return {'aws' : aws_args}

    def parse_scar_args(self, cmd_args):
        scar_args = ['func', 'conf_file', 'json',
                     'verbose', 'path',
                     'preheat', 'execution_mode',
                     'output_file', 'supervisor_version']
        return {'scar' : DataTypesUtils.parse_arg_list(scar_args, cmd_args)}

    def parse_lambda_args(self, cmd_args):
        lambda_args = ['name', 'asynchronous', 'init_script', 'run_script', 'c_args', 'memory', 'time',
                       'timeout_threshold', 'log_level', 'image', 'image_file', 'description',
                       'lambda_role', 'extra_payload', ('environment', 'environment_variables'),
                       'layers', 'lambda_environment', 'list_layers', 'all']
        return DataTypesUtils.parse_arg_list(lambda_args, cmd_args)

    def parse_batch_args(self, cmd_args):
        batch_args = [('batch_vcpus', 'vcpus'), ('batch_memory', 'memory'), 'enable_gpu']
        return DataTypesUtils.parse_arg_list(batch_args, cmd_args)

    def parse_iam_args(self, cmd_args):
        iam_args = [('iam_role', 'role')]
        return DataTypesUtils.parse_arg_list(iam_args, cmd_args)

    def parse_cloudwatchlogs_args(self, cmd_args):
        cw_log_args = ['log_stream_name', 'request_id']
        return DataTypesUtils.parse_arg_list(cw_log_args, cmd_args)

    def parse_api_gateway_args(self, cmd_args):
        api_gtw_args = [('api_gateway_name', 'name'), 'parameters', 'data_binary', 'json_data']
        return DataTypesUtils.parse_arg_list(api_gtw_args, cmd_args)

    def parse_s3_args(self, cmd_args):
        s3_args = ['deployment_bucket',
                   'input_bucket',
                   'output_bucket',
                   ('bucket', 'input_bucket')]
        return DataTypesUtils.parse_arg_list(s3_args, cmd_args)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from enum import Enum
from tabulate import tabulate
import scar.logger as logger
from scar.utils import StrUtils


class OutputType(Enum):
    PLAIN_TEXT = 1
    JSON = 2
    VERBOSE = 3
    BINARY = 4


def parse_http_response(response, function_name, asynch, output_type, output_file):
    if response.ok:
        if output_type == OutputType.BINARY:
            with open(output_file, "wb") as out:
                out.write(StrUtils.decode_base64(response.text))
            text_message = f"Output saved in file '{output_file}'"
        else:
            text_message = f"Request Id: {response.headers['amz-lambda-request-id']}"
            if asynch:
                text_message += f"\nFunction '{function_name}' launched correctly"
            else:
                text_message += f"\nLog Group Name: {response.headers['amz-log-group-name']}\n"
                text_message += f"Log Stream Name: {response.headers['amz-log-stream-name']}\n"
                text_message += json.loads(response.text)["udocker_output"]
    else:
        if asynch and response.status_code == 502:
            text_message = f"Function '{function_name}' launched sucessfully."
        else:
            error = json.loads(response.text)
            if 'message' in error:
                text_message = f"Error ({response.reason}): {error['message']}"
            else:
                text_message = f"Error ({response.reason}): {error['exception']}"
    logger.info(text_message)


def _print_generic_response(response, output_type, aws_output, text_message=None, json_output=None, verbose_output=None, output_file=None):
    if output_type == OutputType.BINARY:
        with open(output_file, "wb") as out:
            out.write(StrUtils.decode_base64(response['Payload']['body']))
    elif output_type == OutputType.PLAIN_TEXT:
        output = text_message
        logger.info(output)
    else:
        if output_type == OutputType.JSON:
            output = json_output if json_output else {aws_output :
                                                      {'RequestId' : response['ResponseMetadata']['RequestId'],
                                                       'HTTPStatusCode' : response['ResponseMetadata']['HTTPStatusCode']}}
        elif output_type == OutputType.VERBOSE:
            output = verbose_output if verbose_output else {aws_output : response}
        logger.info_json(output)


def parse_lambda_function_creation_response(response, function_name, access_key, output_type):
    if response:
        aws_output = 'LambdaOutput'
        text_message = f"Function '{function_name}' successfully created."
        json_message = {aws_output : {'AccessKey' : access_key,
                                      'FunctionArn' : response['FunctionArn'],
                                      'Timeout' : response['Timeout'],
                                      'MemorySize' : response['MemorySize'],
                                      'FunctionName' : response['FunctionName']}}
        _print_generic_response(response, output_type, aws_output, text_message, json_output=json_message)


def parse_log_group_creation_response(response, log_group_name, output_type):
    if response:
        text_message = f"Log group '{log_group_name}' successfully created."
        _print_generic_response(response, output_type, 'CloudWatchOutput', text_message)


def parse_delete_function_response(response, function_name, output_type):
    if response:
        text_message = f"Function '{function_name}' successfully deleted."
        _print_generic_response(response, output_type, 'LambdaOutput', text_message)


def parse_delete_log_response(response, log_group_name, output_type):
    if response:
        text_message = f"Log group '{log_group_name}' successfully deleted."
        _print_generic_response(response, output_type, 'CloudWatchOutput', text_message)


def parse_delete_api_response(response, api_id, output_type):
    if response:
        text_message = f"API Endpoint '{api_id}' successfully deleted."
        _print_generic_response(response, output_type, 'APIGateway', text_message)


def parse_ls_response(lambda_functions, output_type):
    aws_output = 'Functions'
    result = []
    text_message = ""
    if output_type == OutputType.VERBOSE:
        result = lambda_functions
    else:
        for lambdaf in lambda_functions:
            result.append(_parse_lambda_function_info(lambdaf))
        text_message = _get_table(result)
    json_message = { aws_output : result }
    _print_generic_response('', output_type, aws_output, text_message, json_output=json_message, verbose_output=json_message)


def _parse_lambda_function_info(function_info):
    name = function_info.get('FunctionName', "-")
    memory = function_info.get('MemorySize', "-")
    timeout = function_info.get('Timeout', "-")
    image_id = function_info['Environment']['Variables'].get('IMAGE_ID', "-")
    api_gateway = function_info['Environment']['Variables'].get('API_GATEWAY_ID', "-")
    if api_gateway != '-':
        region = function_info['FunctionArn'].split(':')[3]
        api_gateway = f"https://{api_gateway}.execute-api.{region}.amazonaws.com/scar/launch"
    super_version = function_info.get('SupervisorVersion', '-')
    return {'Name' : name,
            'Memory' : memory,
            'Timeout' : timeout,
            'Image_id': image_id,
            'Api_gateway': api_gateway,
            'Sup_version': super_version}


def _get_table(functions_info):
    headers = ['NAME', 'MEMORY', 'TIME', 'IMAGE_ID', 'API_URL', 'SUPERVISOR_VERSION']
    table = []
    for function in functions_info:
        table.append([function['Name'],
                      function['Memory'],
                      function['Timeout'],
                      function['Image_id'],
                      function['Api_gateway'],
                      function['Sup_version']])
    return tabulate(table, headers)


def _parse_error_invocation_response(response, function_name):
    if response:
        if "Task timed out" in response['Payload']:
            # Find the timeout time
            message = StrUtils.find_expression(str(response['Payload']), '(Task timed out .* seconds)')
            # Modify the error message to ease the error readability
            error_msg = message.replace("Task", "Function '%s'" % function_name)
            error_log = f"Error in function response: {error_msg}"
        else:
            error_msg = "Error in function response."
            error_log = f"Error in function response: {response['Payload']}"
        logger.error(error_msg, error_log)


def _parse_payload(value):
    if 'Payload' in value and value['Payload']:
        payload = value['Payload'].read()
        if len(payload) > 0:
            value['Payload'] = json.loads(payload.decode("utf-8"))


def _parse_asynchronous_invocation_response(response, output_type, function_name):
    if response:
        aws_output = 'LambdaOutput'
        text_message = f"Request Id: {response['ResponseMetadata']['RequestId']}\n"
        text_message += f"Function '{function_name}' launched correctly"
        json_message = {aws_output : {'StatusCode' : response['StatusCode'],
                                       'RequestId' : response['ResponseMetadata']['RequestId']}}
        _print_generic_response(response, output_type, aws_output, text_message, json_output=json_message)


def _parse_requestresponse_invocation_response(**kwargs):
    if kwargs['Response']:
        response = kwargs['Response']
        aws_output = 'LambdaOutput'
        log_group_name = response['Payload']['headers']['amz-log-group-name']
        log_stream_name = response['Payload']['headers']['amz-log-stream-name']
        request_id = response['ResponseMetadata']['RequestId']
        if "exception" in response['Payload']['body']:
            body = ("ERROR launching udocker container: \n "
                    f"{json.loads(response['Payload']['body'])['exception']}")
        else:
            body = StrUtils.base64_to_utf8_string(response['Payload']['body'])

        text_message = (f"Request Id: {request_id}\n"
                        f"Log Group Name: {log_group_name}\n"
                        f"Log Stream Name: {log_stream_name}\n")
        text_message += body

        json_message = {aws_output : {'StatusCode' : response['StatusCode'],
                                      'Payload' : body,
                                      'LogGroupName' : log_group_name,
                                      'LogStreamName' : log_stream_name,
                                      'RequestId' : request_id}}
        if 'OutputFile' in kwargs and kwargs['OutputFile']:
            _print_generic_response(response, kwargs['OutputType'], aws_output, text_message, json_output=json_message, output_file=kwargs['OutputFile'])
        else:
            _print_generic_response(response, kwargs['OutputType'], aws_output, text_message, json_output=json_message)


def _parse_base64_response_values(value):
    value['LogResult'] = StrUtils.base64_to_utf8_string(value['LogResult'])
    value['ResponseMetadata']['HTTPHeaders']['x-amz-log-result'] = \
        StrUtils.base64_to_utf8_string(value['ResponseMetadata']['HTTPHeaders']['x-amz-log-result'])


def parse_invocation_response(**kwargs):
    # Decode and parse the payload
    _parse_payload(kwargs['Response'])
    if "FunctionError" in kwargs['Response']:
        _parse_error_invocation_response(kwargs['Response'], kwargs['FunctionName'])
    if kwargs['IsAsynchronous']:
        _parse_asynchronous_invocation_response(kwargs['Response'], kwargs['OutputType'], kwargs['FunctionName'])
    else:
        # Transform the base64 encoded results to something legible
        _parse_base64_response_values(kwargs['Response'])
        # Extract log_group_name and log_stream_name from the payload
        _parse_requestresponse_invocation_response(**kwargs)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods used to manage the AWS tools."""

import os
from typing import Dict
from scar.cmdtemplate import Commands
from scar.providers.aws.apigateway import APIGateway
from scar.providers.aws.batchfunction import Batch
from scar.providers.aws.cloudwatchlogs import CloudWatchLogs
from scar.providers.aws.iam import IAM
from scar.providers.aws.lambdafunction import Lambda
from scar.providers.aws.properties import AwsProperties, ScarProperties
from scar.providers.aws.resourcegroups import ResourceGroups
from scar.providers.aws.s3 import S3
from scar.providers.aws.validators import AWSValidator
from scar.providers.aws.properties import ApiGatewayProperties
import scar.exceptions as excp
import scar.logger as logger
import scar.providers.aws.response as response_parser
from scar.utils import lazy_property, StrUtils, FileUtils

_ACCOUNT_ID_REGEX = r'\d{12}'


def _get_storage_provider_id(storage_provider: str, env_vars: Dict) -> str:
    """Searches the storage provider id in the environment variables:
        get_provider_id(S3, {'STORAGE_AUTH_S3_USER_41807' : 'scar'})
        returns -> 41807"""
    res = ""
    for env_key in env_vars.keys():
        if env_key.startswith(f'STORAGE_AUTH_{storage_provider}'):
            res = env_key.split('_', 4)[-1]
            break
    return res


class AWS(Commands):
    """AWS controller.
    Used to manage all the AWS calls and functionalities."""

    @lazy_property
    def aws_lambda(self):
        """It's called 'aws_lambda' because 'lambda'
        it's a restricted word in python."""
        aws_lambda = Lambda(self.aws_properties,
                            self.scar_properties.supervisor_version)
        return aws_lambda

    @lazy_property
    def batch(self):
        batch = Batch(self.aws_properties,
                      self.scar_properties.supervisor_version)
        return batch

    @lazy_property
    def cloudwatch_logs(self):
        cloudwatch_logs = CloudWatchLogs(self.aws_properties)
        return cloudwatch_logs

    @lazy_property
    def api_gateway(self):
        api_gateway = APIGateway(self.aws_properties)
        return api_gateway

    @lazy_property
    def aws_s3(self):
        aws_s3 = S3(self.aws_properties)
        return aws_s3

    @lazy_property
    def resource_groups(self):
        resource_groups = ResourceGroups(self.aws_properties)
        return resource_groups

    @lazy_property
    def iam(self):
        iam = IAM(self.aws_properties)
        return iam

    @excp.exception(logger)
    def init(self):
        if self.aws_lambda.find_function():
            raise excp.FunctionExistsError(function_name=self.aws_properties.lambdaf.name)
        # We have to create the gateway before creating the function
        self._create_api_gateway()
        self._create_lambda_function()
        self._create_log_group()
        self._create_s3_buckets()
        # The api_gateway permissions are added after the function is created
        self._add_api_gateway_permissions()
        self._create_batch_environment()
        self._preheat_function()

    @excp.exception(logger)
    def invoke(self):
        self._update_local_function_properties()
        response = self.aws_lambda.call_http_endpoint()
        response_parser.parse_http_response(response,
                                            self.aws_properties.lambdaf.name,
                                            self.aws_properties.lambdaf.asynchronous,
                                            self.aws_properties.output,
                                            getattr(self.scar_properties, "output_file", ""))

    @excp.exception(logger)
    def run(self):
        if hasattr(self.aws_properties, "s3") and hasattr(self.aws_properties.s3, "input_bucket"):
            self._process_input_bucket_calls()
        else:
            if self.aws_lambda.is_asynchronous():
                self.aws_lambda.set_asynchronous_call_parameters()
            self.aws_lambda.launch_lambda_instance()

    @excp.exception(logger)
    def update(self):
        if hasattr(self.aws_properties.lambdaf, "all") and self.aws_properties.lambdaf.all:
            self._update_all_functions(self._get_all_functions())
        else:
            self.aws_lambda.update_function_configuration()

    @excp.exception(logger)
    def ls(self):
        if hasattr(self.aws_properties, "s3"):
            file_list = self.aws_s3.get_bucket_file_list()
            for file_info in file_list:
                print(file_info)
        else:
            lambda_functions = self._get_all_functions()
            response_parser.parse_ls_response(lambda_functions,
                                              self.aws_properties.output)

    @excp.exception(logger)
    def rm(self):
        if hasattr(self.aws_properties.lambdaf, "all") and self.aws_properties.lambdaf.all:
            self._delete_all_resources()
        else:
            function_info = self.aws_lambda.get_function_info(self.aws_properties.lambdaf.name)
            self._delete_resources(function_info)

    @excp.exception(logger)
    def log(self):
        aws_log = self.cloudwatch_logs.get_aws_log()
        batch_logs = self._get_batch_logs()
        aws_log += batch_logs if batch_logs else ""
        print(aws_log)

    @excp.exception(logger)
    def put(self):
        self.upload_file_or_folder_to_s3()

    @excp.exception(logger)
    def get(self):
        self.download_file_or_folder_from_s3()

    @AWSValidator.validate()
    @excp.exception(logger)
    def parse_arguments(self, **kwargs):
        self.raw_kwargs = kwargs
        self.aws_properties = AwsProperties(kwargs.get('aws', {}))
        self.scar_properties = ScarProperties(kwargs.get('scar', {}))
        self.add_extra_aws_properties()

    def add_extra_aws_properties(self):
        self._add_tags()
        self._add_output()
        self._add_account_id()
        self._add_config_file_path()

    def _add_tags(self):
        self.aws_properties.tags = {"createdby": "scar", "owner": self.iam.get_user_name_or_id()}

    def _add_output(self):
        self.aws_properties.output = response_parser.OutputType.PLAIN_TEXT
        if hasattr(self.scar_properties, "json") and self.scar_properties.json:
            self.aws_properties.output = response_parser.OutputType.JSON
        # Override json ouput if both of them are defined
        if hasattr(self.scar_properties, "verbose") and self.scar_properties.verbose:
            self.aws_properties.output = response_parser.OutputType.VERBOSE
        if hasattr(self.scar_properties, "output_file") and self.scar_properties.output_file:
            self.aws_properties.output = response_parser.OutputType.BINARY
            self.aws_properties.output_file = self.scar_properties.output_file

    def _add_account_id(self):
        self.aws_properties.account_id = StrUtils.find_expression(self.aws_properties.iam.role,
                                                                  _ACCOUNT_ID_REGEX)

    def _add_config_file_path(self):
        if hasattr(self.scar_properties, "conf_file") and self.scar_properties.conf_file:
            self.aws_properties.config_path = os.path.dirname(self.scar_properties.conf_file)

    def _get_all_functions(self):
        arn_list = self.resource_groups.get_resource_arn_list(self.iam.get_user_name_or_id())
        return self.aws_lambda.get_all_functions(arn_list)

    def _get_batch_logs(self) -> str:
        logs = ""
        if hasattr(self.aws_properties.cloudwatch, "request_id") and \
        self.batch.exist_job(self.aws_properties.cloudwatch.request_id):
            batch_jobs = self.batch.describe_jobs(self.aws_properties.cloudwatch.request_id)
            logs = self.cloudwatch_logs.get_batch_job_log(batch_jobs["jobs"])
        return logs

    @excp.exception(logger)
    def _create_lambda_function(self):
        response = self.aws_lambda.create_function()
        acc_key = self.aws_lambda.client.get_access_key()
        response_parser.parse_lambda_function_creation_response(response,
                                                                self.aws_properties.lambdaf.name,
                                                                acc_key,
                                                                self.aws_properties.output)

    @excp.exception(logger)
    def _create_log_group(self):
        response = self.cloudwatch_logs.create_log_group()
        response_parser.parse_log_group_creation_response(response,
                                                          self.cloudwatch_logs.get_log_group_name(),
                                                          self.aws_properties.output)

    @excp.exception(logger)
    def _create_s3_buckets(self):
        if hasattr(self.aws_properties, "s3"):
            if hasattr(self.aws_properties.s3, "input_bucket"):
                self.aws_s3.create_input_bucket(create_input_folder=True)
                self.aws_lambda.link_function_and_input_bucket()
                self.aws_s3.set_input_bucket_notification()
            if hasattr(self.aws_properties.s3, "output_bucket"):
                self.aws_s3.create_output_bucket()

    def _create_api_gateway(self):
        if hasattr(self.aws_properties, "api_gateway"):
            self.api_gateway.create_api_gateway()

    def _add_api_gateway_permissions(self):
        if hasattr(self.aws_properties, "api_gateway"):
            self.aws_lambda.add_invocation_permission_from_api_gateway()

    def _create_batch_environment(self):
        if self.aws_properties.execution_mode == "batch" or \
        self.aws_properties.execution_mode == "lambda-batch":
            self.batch.create_batch_environment()

    def _preheat_function(self):
        # If preheat is activated, the function is launched at the init step
        if hasattr(self.scar_properties, "preheat"):
            self.aws_lambda.preheat_function()

    def _process_input_bucket_calls(self):
        s3_file_list = self.aws_s3.get_bucket_file_list()
        logger.info(f"Files found: '{s3_file_list}'")
        # First do a request response invocation to prepare the lambda environment
        if s3_file_list:
            s3_event = self.aws_s3.get_s3_event(s3_file_list.pop(0))
            self.aws_lambda.launch_request_response_event(s3_event)
        # If the list has more elements, invoke functions asynchronously
        if s3_file_list:
            s3_event_list = self.aws_s3.get_s3_event_list(s3_file_list)
            self.aws_lambda.process_asynchronous_lambda_invocations(s3_event_list)

    def upload_file_or_folder_to_s3(self):
        path_to_upload = self.scar_properties.path
        self.aws_s3.create_input_bucket()
        files = [path_to_upload]
        if os.path.isdir(path_to_upload):
            files = FileUtils.get_all_files_in_directory(path_to_upload)
        for file_path in files:
            self.aws_s3.upload_file(folder_name=self.aws_properties.s3.input_folder,
                                    file_path=file_path)

    def _get_download_file_path(self, file_key=None):
        file_path = file_key
        if hasattr(self.scar_properties, "path") and self.scar_properties.path:
            file_path = FileUtils.join_paths(self.scar_properties.path, file_path)
        return file_path

    def download_file_or_folder_from_s3(self):
        bucket_name = self.aws_properties.s3.input_bucket
        s3_file_list = self.aws_s3.get_bucket_file_list()
        for s3_file in s3_file_list:
            # Avoid download s3 'folders'
            if not s3_file.endswith('/'):
                file_path = self._get_download_file_path(file_key=s3_file)
                # make sure the path folders are created
                dir_path = os.path.dirname(file_path)
                if dir_path and not os.path.isdir(dir_path):
                    os.makedirs(dir_path, exist_ok=True)
                self.aws_s3.download_file(bucket_name, s3_file, file_path)

    def _update_all_functions(self, lambda_functions):
        for function_info in lambda_functions:
            self.aws_lambda.update_function_configuration(function_info)

    def _update_local_function_properties(self, function_info):
        self._reset_aws_properties()
        """Update the defined properties with the AWS information."""
        if function_info:
            self.aws_properties.lambdaf.update_properties(**function_info)
        if 'API_GATEWAY_ID' in self.aws_properties.lambdaf.environment['Variables']:
            api_gtw_id = self.aws_properties.lambdaf.environment['Variables'].get('API_GATEWAY_ID',
                                                                                  "")
            if hasattr(self.aws_properties, 'api_gateway'):
                self.aws_properties.api_gateway.id = api_gtw_id
            else:
                self.aws_properties.api_gateway = ApiGatewayProperties({'id' : api_gtw_id})

#############################################################################
###                   Methods to delete AWS resources                     ###
#############################################################################

    def _delete_all_resources(self):
        for function_info in self._get_all_functions():
            self._delete_resources(function_info)

    def _delete_resources(self, function_info):
        function_name = function_info['FunctionName']
        if not self.aws_lambda.find_function(function_name):
            raise excp.FunctionNotFoundError(function_name=function_name)
        # Delete associated api
        self._delete_api_gateway(function_info['Environment']['Variables'])
        # Delete associated log
        self._delete_logs(function_name)
        # Delete associated notifications
        self._delete_bucket_notifications(function_info['FunctionArn'],
                                          function_info['Environment']['Variables'])
        # Delete function
        self._delete_lambda_function(function_name)
        # Delete resources batch
        self._delete_batch_resources(function_name)

    def _delete_api_gateway(self, function_env_vars):
        api_gateway_id = function_env_vars.get('API_GATEWAY_ID')
        if api_gateway_id:
            response = self.api_gateway.delete_api_gateway(api_gateway_id)
            response_parser.parse_delete_api_response(response, api_gateway_id,
                                                      self.aws_properties.output)

    def _delete_logs(self, function_name):
        log_group_name = self.cloudwatch_logs.get_log_group_name(function_name)
        response = self.cloudwatch_logs.delete_log_group(log_group_name)
        response_parser.parse_delete_log_response(response,
                                                  log_group_name,
                                                  self.aws_properties.output)

    def _delete_bucket_notifications(self, function_arn, function_env_vars):
        s3_provider_id = _get_storage_provider_id('S3', function_env_vars)
        input_bucket_id = f'STORAGE_PATH_INPUT_{s3_provider_id}' if s3_provider_id else ''
        if input_bucket_id in function_env_vars:
            input_path = function_env_vars[input_bucket_id]
            input_bucket_name = input_path.split("/", 1)[0]
            self.aws_s3.delete_bucket_notification(input_bucket_name, function_arn)

    def _delete_lambda_function(self, function_name):
        response = self.aws_lambda.delete_function(function_name)
        response_parser.parse_delete_function_response(response, function_name,
                                                       self.aws_properties.output)

    def _delete_batch_resources(self, function_name):
        if self.batch.exist_compute_environments(function_name):
            self.batch.delete_compute_environment(function_name)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes to create the function deployment package."""

from zipfile import ZipFile
from scar.providers.aws.udocker import Udocker
from scar.providers.aws.validators import AWSValidator
from scar.exceptions import exception
import scar.logger as logger
from scar.http.request import get_file
from scar.utils import FileUtils, lazy_property, GitHubUtils, \
                       GITHUB_USER, GITHUB_SUPERVISOR_PROJECT

_SUPERVISOR_ZIP_PATH = FileUtils.join_paths(FileUtils.get_tmp_dir(), 'faas.zip')


def _download_faas_supervisor_zip(supervisor_version: str) -> None:
    supervisor_zip_url = GitHubUtils.get_source_code_url(GITHUB_USER, GITHUB_SUPERVISOR_PROJECT,
                                                         supervisor_version)
    with open(_SUPERVISOR_ZIP_PATH, "wb") as thezip:
        thezip.write(get_file(supervisor_zip_url))


def _extract_handler_code(scar_tmp_folder_path: str, handler_name: str) -> None:
    function_handler_dest = FileUtils.join_paths(scar_tmp_folder_path, f"{handler_name}.py")
    file_path = ""
    with ZipFile(_SUPERVISOR_ZIP_PATH) as thezip:
        for file in thezip.namelist():
            if file.endswith("function_handler.py"):
                file_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), file)
                thezip.extract(file, FileUtils.get_tmp_dir())
                break
    FileUtils.copy_file(file_path, function_handler_dest)


_INIT_SCRIPT_NAME = "init_script.sh"


class FunctionPackager():
    """Class to manage the deployment package creation."""

    @lazy_property
    def udocker(self):
        """Udocker client"""
        udocker = Udocker(self.aws, self.scar_tmp_folder_path, _SUPERVISOR_ZIP_PATH)
        return udocker

    def __init__(self, aws_properties, supervisor_version):
        self.aws = aws_properties
        self.supervisor_version = supervisor_version
        self.scar_tmp_folder = FileUtils.create_tmp_dir()
        self.scar_tmp_folder_path = self.scar_tmp_folder.name

        self.package_args = {}

    @exception(logger)
    def create_zip(self):
        """Creates the lambda function deployment package."""
        self._clean_tmp_folders()
        _download_faas_supervisor_zip(self.supervisor_version)
        _extract_handler_code(self.scar_tmp_folder_path,
                              self.aws.lambdaf.name)
        self._manage_udocker_images()
        self._add_init_script()
        self._add_extra_payload()
        self._zip_scar_folder()
        self._check_code_size()
        # self._clean_tmp_folders()

    def _clean_tmp_folders(self):
        FileUtils.delete_file(_SUPERVISOR_ZIP_PATH)
        FileUtils.delete_file(self.aws.lambdaf.zip_file_path)

    def _manage_udocker_images(self):
        if hasattr(self.aws.lambdaf, "image") and \
           hasattr(self.aws, "s3") and \
           hasattr(self.aws.s3, "deployment_bucket"):
            self.udocker.download_udocker_image()
        if hasattr(self.aws.lambdaf, "image_file"):
            if hasattr(self.aws, "config_path"):
                self.aws.lambdaf.image_file = FileUtils.join_paths(self.aws.config_path,
                                                                   self.aws.lambdaf.image_file)
            self.udocker.prepare_udocker_image()

    def _add_init_script(self):
        if hasattr(self.aws.lambdaf, "init_script"):
            if hasattr(self.aws, "config_path"):
                self.aws.lambdaf.init_script = FileUtils.join_paths(self.aws.config_path,
                                                                    self.aws.lambdaf.init_script)
            FileUtils.copy_file(self.aws.lambdaf.init_script,
                                FileUtils.join_paths(self.scar_tmp_folder_path, _INIT_SCRIPT_NAME))
            self.aws.lambdaf.environment['Variables']['INIT_SCRIPT_PATH'] = \
                                        f"/var/task/{_INIT_SCRIPT_NAME}"

    def _add_extra_payload(self):
        if hasattr(self.aws.lambdaf, "extra_payload"):
            logger.info("Adding extra payload from {0}".format(self.aws.lambdaf.extra_payload))
            FileUtils.copy_dir(self.aws.lambdaf.extra_payload, self.scar_tmp_folder_path)
            self.aws.lambdaf.environment['Variables']['EXTRA_PAYLOAD'] = "/var/task"

    def _zip_scar_folder(self):
        FileUtils.zip_folder(self.aws.lambdaf.zip_file_path,
                             self.scar_tmp_folder_path,
                             "Creating function package")

    def _check_code_size(self):
        # Check if the code size fits within the AWS limits
        if hasattr(self.aws, "s3") and hasattr(self.aws.s3, "deployment_bucket"):
            AWSValidator.validate_s3_code_size(self.scar_tmp_folder_path,
                                               self.aws.lambdaf.max_s3_payload_size)
        else:
            AWSValidator.validate_function_code_size(self.scar_tmp_folder_path,
                                                     self.aws.lambdaf.max_payload_size)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods to manage the different
properties needed by SCAR and the boto clients."""


class ScarProperties(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class AwsProperties(dict):

    def __init__(self, *args, **kwargs):
        """
        {'account_id': '914332',
         'batch': see_batch_props_class,
         'boto_profile': 'default',
         'cloudwatch': see_cloudwatch_props_class,
         'config_path': 'cowsay',
         'execution_mode': 'lambda',
         'iam': see_iam_props_class,
         'lambda': see_lambdaf_props_class,
         'output': <OutputType.PLAIN_TEXT: 1>,
         'region': 'us-east-1',
         's3': see_s3_props_class,
         'tags': {'createdby': 'scar', 'owner': 'alpegon'}}
        """
        super().__init__(*args, **kwargs)
        self.__dict__ = self
        self._initialize_properties()

    def _initialize_properties(self):
        if hasattr(self, "api_gateway"):
            self.api_gateway = ApiGatewayProperties(self.api_gateway)
        if hasattr(self, "batch"):
            self.batch = BatchProperties(self.batch)
        if hasattr(self, "cloudwatch"):
            self.cloudwatch = CloudWatchProperties(self.cloudwatch)
        if hasattr(self, "iam"):
            self.iam = IamProperties(self.iam)
        if hasattr(self, "lambda"):
            self.lambdaf = LambdaProperties(self.__dict__['lambda'])
            self.__dict__.pop('lambda', None)
        if hasattr(self, "s3"):
            self.s3 = S3Properties(self.s3)


class ApiGatewayProperties(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class BatchProperties(dict):
    """
    Example of dictionary used to initialize the class properties:
    {'comp_type': 'EC2',
    'desired_v_cpus': 0,
    'instance_types': ['m3.medium'],
    'max_v_cpus': 2,
    'min_v_cpus': 0,
    'security_group_ids': ['sg-2568'],
    'state': 'ENABLED',
    'subnets': ['subnet-568',
                'subnet-569',
                'subnet-570',
                'subnet-571',
                'subnet-572'],
    'type': 'MANAGED'}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class LambdaProperties(dict):
    """
    Example of dictionary used to initialize the class properties:
    {'asynchronous': False,
    'description': 'Automatically generated lambda function',
    'environment': {'Variables': {'EXECUTION_MODE': 'lambda',
                                  'INPUT_BUCKET': 'test1',
                                  'LOG_LEVEL': 'INFO',
                                  'SUPERVISOR_TYPE': 'LAMBDA',
                                  'TIMEOUT_THRESHOLD': '10',
                                  'UDOCKER_BIN': '/opt/udocker/bin/',
                                  'UDOCKER_DIR': '/tmp/shared/udocker',
                                  'UDOCKER_EXEC': '/opt/udocker/udocker.py',
                                  'UDOCKER_LIB': '/opt/udocker/lib/'}},
    'extra_payload': '/test/',
    'handler': 'test.lambda_handler',
    'image_file': 'minicow.tar.gz',
    'init_script': 'test.sh',
    'invocation_type': 'RequestResponse',
    'layers': ['arn:aws:lambda:us-east-1:914332:layer:faas-supervisor:1'],
    'log_level': 'INFO',
    'log_type': 'Tail',
    'memory': 512,
    'name': 'test',
    'runtime': 'python3.6',
    'time': 300,
    'timeout_threshold': 10,
    'zip_file_path': '/tmp/function.zip'}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def update_properties(self, **kwargs):
        if 'ResponseMetadata' in kwargs:
            # Parsing RAW function info
            self.description = kwargs['Description']
            self.environment = kwargs['Environment']
            self.arn = kwargs['FunctionArn']
            self.name = kwargs['FunctionName']
            self.handler = kwargs['Handler']
            self.layers = [layer['Arn'] for layer in kwargs['Layers']]
            self.memory = kwargs['MemorySize']
            self.time = kwargs['Timeout']
            self.role = kwargs['Role']
            self.runtime = kwargs['Runtime']
        else:
            self.__dict__.update(**kwargs)


class IamProperties(dict):
    """
    Example of dictionary used to initialize the class properties:
    {'role': 'arn:aws:iam::914332:role/invented-role'}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class S3Properties(dict):
    """
    Example of dictionary used to initialize the class properties:
    {'input_bucket': 'test1'}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self
        self.process_storage_paths()

    def process_storage_paths(self):
        if hasattr(self, "input_bucket"):
            self.storage_path_input = self.input_bucket
            input_path = self.input_bucket.split("/", 1)
            if len(input_path) > 1:
                # There are folders defined
                self.input_bucket = input_path[0]
                self.input_folder = input_path[1]

        if hasattr(self, "output_bucket"):
            self.storage_path_output = self.output_bucket
            output_path = self.output_bucket.split("/", 1)
            if len(output_path) > 1:
                # There are folders defined
                self.output_bucket = output_path[0]
                self.output_folder = output_path[1]


class CloudWatchProperties(dict):
    """
    Example of dictionary used to initialize the class properties:
    {'log_retention_policy_in_days': 30}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes to manage the Lambda layers."""

import io
import shutil
from typing import Dict
import zipfile
from tabulate import tabulate
import scar.http.request as request
import scar.logger as logger
from scar.utils import lazy_property, FileUtils, GitHubUtils, StrUtils, \
                       GITHUB_USER, GITHUB_SUPERVISOR_PROJECT


def _create_tmp_folders() -> None:
    tmp_zip_folder = FileUtils.create_tmp_dir()
    layer_code_folder = FileUtils.create_tmp_dir()
    return (tmp_zip_folder.name, layer_code_folder.name)


def _download_supervisor(supervisor_version: str, tmp_zip_path: str) -> str:
    supervisor_zip_url = GitHubUtils.get_source_code_url(GITHUB_USER, GITHUB_SUPERVISOR_PROJECT,
                                                         supervisor_version)
    supervisor_zip = request.get_file(supervisor_zip_url)
    with zipfile.ZipFile(io.BytesIO(supervisor_zip)) as thezip:
        for file in thezip.namelist():
            # Remove the parent folder path
            parent_folder, file_name = file.split("/", 1)
            if file_name.startswith("extra") or file_name.startswith("faassupervisor"):
                thezip.extract(file, tmp_zip_path)
    return parent_folder


def _copy_supervisor_files(parent_folder: str, tmp_zip_path: str, layer_code_path: str) -> None:
    supervisor_path = FileUtils.join_paths(tmp_zip_path, parent_folder, 'faassupervisor')
    shutil.move(supervisor_path, FileUtils.join_paths(layer_code_path, 'python', 'faassupervisor'))


def _copy_extra_files(parent_folder: str, tmp_zip_path: str, layer_code_path: str) -> None:
    extra_folder_path = FileUtils.join_paths(tmp_zip_path, parent_folder, 'extra')
    files = FileUtils.get_all_files_in_directory(extra_folder_path)
    for file_path in files:
        FileUtils.unzip_folder(file_path, layer_code_path)


def _create_layer_zip(layer_zip_path: str, layer_code_path: str) -> None:
    FileUtils.zip_folder(layer_zip_path, layer_code_path)


class Layer():
    """Class used for layer management."""

    def __init__(self, lambda_client) -> None:
        self.lambda_client = lambda_client

    def _find(self, layer_name: str) -> Dict:
        """Returns the layer information that matches the layer name passed."""
        for layer in self.lambda_client.list_layers():
            if layer.get('LayerName', '') == layer_name:
                return layer
        return {}

    def create(self, **kwargs: Dict) -> Dict:
        """Creates a new layer with the arguments passed."""
        return self.lambda_client.publish_layer_version(**kwargs)

    def exists(self, layer_name: str) -> bool:
        """Checks layer name for existence."""
        if self._find(layer_name):
            return True
        return False

    def delete(self, **kwargs: Dict) -> Dict:
        """Deletes a layer."""
        layer_args = {'LayerName' : kwargs['name']}
        if 'version' in kwargs:
            layer_args['VersionNumber'] = int(kwargs['version'])
        else:
            version_info = self.get_latest_layer_info(kwargs['name'])
            layer_args['VersionNumber'] = version_info.get('Version', -1)
        return self.lambda_client.delete_layer_version(**layer_args)

    def get_latest_layer_info(self, layer_name: str) -> str:
        """Returns the latest matching version of the layer with 'layer_name'."""
        layer = self._find(layer_name)
        return layer['LatestMatchingVersion'] if layer else {}


class LambdaLayers():
    """"Class used to manage the lambda supervisor layer."""

    _SUPERVISOR_LAYER_NAME = 'faas-supervisor'

    @lazy_property
    def layer(self):
        """Property used to manage the lambda layers."""
        layer = Layer(self.lambda_client)
        return layer

    def __init__(self, lambda_client, supervisor_version: str) -> None:
        self.lambda_client = lambda_client
        self.supervisor_version = supervisor_version

    def _get_supervisor_layer_props(self, layer_zip_path: str) -> Dict:
        return {'LayerName' : self._SUPERVISOR_LAYER_NAME,
                'Description' : self.supervisor_version,
                'Content' : {'ZipFile': FileUtils.read_file(layer_zip_path, mode="rb")},
                'LicenseInfo' : 'Apache 2.0'}

    def _create_layer(self) -> None:
        tmp_zip_path, layer_code_path = _create_tmp_folders()
        layer_zip_path = FileUtils.join_paths(FileUtils.get_tmp_dir(),
                                              f"{self._SUPERVISOR_LAYER_NAME}.zip")
        parent_folder = _download_supervisor(self.supervisor_version, tmp_zip_path)
        _copy_supervisor_files(parent_folder, tmp_zip_path, layer_code_path)
        _copy_extra_files(parent_folder, tmp_zip_path, layer_code_path)
        _create_layer_zip(layer_zip_path, layer_code_path)
        self.layer.create(**self._get_supervisor_layer_props(layer_zip_path))
        FileUtils.delete_file(layer_zip_path)

    def _create_supervisor_layer(self) -> None:
        logger.info("Creating faas-supervisor layer.")
        self._create_layer()
        logger.info("Faas-supervisor layer created.")

    def _update_supervisor_layer(self) -> None:
        logger.info("Updating faas-supervisor layer.")
        self._create_layer()
        logger.info("Faas-supervisor layer updated.")

    def print_layers_info(self) -> None:
        """Prints the lambda layers information."""
        layers_info = self.lambda_client.list_layers()
        headers = ['NAME', 'VERSION', 'ARN', 'RUNTIMES']
        table = []
        for layer in layers_info:
            table.append([layer.get('LayerName', ""),
                          layer.get('LatestMatchingVersion', {}).get('Version', -1),
                          layer.get('LayerArn', ""),
                          layer.get('LatestMatchingVersion', {}).get('CompatibleRuntimes', '-')])
        print(tabulate(table, headers))

    def get_latest_supervisor_layer_arn(self) -> str:
        """Returns the ARN of the latest supervisor layer."""
        layer_info = self.layer.get_latest_layer_info(self._SUPERVISOR_LAYER_NAME)
        return layer_info.get('LayerVersionArn', "")

    def check_faas_supervisor_layer(self):
        """Checks if the supervisor layer exists, if not, creates the layer.
        If the layer exists and it's not updated, updates the layer."""
        # Get the layer information
        layer_info = self.layer.get_latest_layer_info(self._SUPERVISOR_LAYER_NAME)
        # Compare supervisor versions
        if layer_info and 'Description' in layer_info:
            # If the supervisor layer version is lower than the passed version,
            # we must update the layer
            if StrUtils.compare_versions(layer_info.get('Description', ''),
                                         self.supervisor_version) < 0:
                self._update_supervisor_layer()
            else:
                logger.info("Using existent 'faas-supervisor' layer")
        else:
            # Layer not found, we have to create it
            self._create_supervisor_layer()
